from __future__ import absolute_import

from jira import JIRA, JIRAError

from hamster_bridge.listeners import HamsterListener

import logging
import re

logger = logging.getLogger(__name__)


class JiraHamsterListener(HamsterListener):

    short_name = 'jira'

    config_values = [
        ('server_url', lambda: raw_input('Root url to your jira server [f.e. "http://jira.example.org"]\n')),
        ('username', lambda: raw_input('Your jira user name\n')),
        ('password', lambda: raw_input('Your jira password\n')),
        ('auto_start', lambda: raw_input('Automatically start the issue when you start the task in hamster? [y/n]\n'))
    ]

    issue_from_title = re.compile('([A-Z][A-Z0-9]+-[0-9]+)')

    # noinspection PyBroadException
    def prepare(self):
        server_url = self.config.get(self.short_name, 'server_url')
        username = self.config.get(self.short_name, 'username')
        password = self.config.get(self.short_name, 'password')

        logger.info('Connecting as "%s" to "%s"', username, server_url)
        self.jira = JIRA(server_url, basic_auth=(username, password))

        try:
            self.jira.projects()
        except:
            logger.exception('Can not connect to JIRA, please check ~/.hamster-bridge.cfg')

    def __issue_from_fact(self, fact):
        """
        Get the issue name from a fact
        :param fact: the fact to search the issue in
        """
        fields = [fact.activity] + fact.tags
        logger.debug('Searching ticket in: %r', fields)
        for field in fields:
            for possible_issue in self.issue_from_title.findall(field):
                logger.debug('Lookup issue for activity "%s"', possible_issue)
                try:
                    self.jira.issue(possible_issue)
                    logger.debug('Found existing issue "%s" in "%s"', possible_issue, field)
                    return possible_issue
                except JIRAError, e:
                    if e.text == 'Issue Does Not Exist':
                        logger.warning('Tried issue "%s", but does not exist. ', possible_issue)
                    else:
                        logger.exception('Error communicating with Jira')

    def on_fact_started(self, fact):
        if self.config.get(self.short_name, 'auto_start') == 'y':
            try:
                issue_name = self.__issue_from_fact(fact)
                if issue_name is None:
                    return

                for transition in self.jira.transitions(issue_name):
                    if transition['name'] == u'Start Progress':
                        self.jira.transition_issue(issue_name, transition['id'])
                        logger.info('Marked issue "%s" as "In Progress"', issue_name)
            except JIRAError:
                logger.exception('Error communicating with Jira')

    def on_fact_stopped(self, fact):
        time_spent = '%dm' % (fact.delta.total_seconds() / 60)
        issue_name = self.__issue_from_fact(fact)
        if issue_name:
            try:
                worklog = self.jira.add_worklog(issue_name, time_spent, comment=fact.description)
                logger.info('Logged work: %s to %s (created %r)', time_spent, issue_name, worklog)
            except JIRAError:
                logger.exception('Error communicating with Jira')
        else:
            logger.debug('No jira issue found')



import argparse
import logging

from hamster_bridge.bridge import HamsterBridge


CONFIG_PATH = '~/.hamster-bridge.cfg'


def import_listener(name):
    components = name.rsplit('.')
    mod = __import__(name.rsplit('.', 1)[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


LISTENERS = [
    'hamster_bridge.listeners.jira.JiraHamsterListener',
    'hamster_bridge.listeners.redmine.RedmineHamsterListener',
    # your backend missing? contributions welcome!
]


def main():
    listener_choices = dict(
        (lc.short_name, lc) for lc in [import_listener(l) for l in LISTENERS]
    )

    parser = argparse.ArgumentParser(description='Let your hamster log your work to your favorite bugtracker.')
    parser.add_argument('bugtracker', choices=sorted(listener_choices.keys()))
    parser.add_argument('-d', '--debug', action='store_true', help='enable debug logging')
    parser.add_argument('-c', '--check-interval', default='2', type=int,
                        help='check every this amount of seconds for updates')
    parser.add_argument('--config-path', default=CONFIG_PATH, type=str, 
                        help='path to config file, defaults to {}'.format(CONFIG_PATH))
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format='%(asctime)-15s %(levelname)+7s{}: %(message)s'.format(' [%(name)s]' if args.debug else '')
    )
    logger = logging.getLogger(__name__)

    logger.info('Starting hamster bridge')
    bridge = HamsterBridge()
    logger.debug('Activating listener: %s', args.bugtracker)
    bridge.add_listener(listener_choices[args.bugtracker]())
    bridge.configure(args.config_path)
    logger.debug('Run with check interval of %ds', args.check_interval)
    bridge.run(args.check_interval)

if __name__ == "__main__":
    main()



import ConfigParser
import datetime
import time
import logging
import os
import stat

logger = logging.getLogger(__name__)


try:
    import hamster.client
except ImportError:
    raise ImportError('Can not find hamster')


class HamsterBridge(hamster.client.Storage):
    """
    Connects to the running hamster instance via dbus. But as the notification does not work reliable there is a
    polling-based loop in the run()-method that will trigger all registered listeners.
    """
    def __init__(self):
        super(HamsterBridge, self).__init__()
        self._listeners = []

    def add_listener(self, listener):
        """
        Registers the given HamsterListener instance. It will then be notified about changes.

        :param listener: the HamsterListener instance
        :type  listener: HamsterListener
        """
        if listener not in self._listeners:
            self._listeners.append(listener)

    def configure(self, config_path):
        """
        Gives each listener the chance to do something before we start the bridge's runtime loops.
        :param config_path: path to config file
        :type config_path:  str
        """
        path = os.path.expanduser(config_path)
        config = ConfigParser.RawConfigParser()
        # read from file if exists
        if os.path.exists(path):
            logger.debug('Reading config file from %s', path)
            config.read(path)
        # let listeners extend
        for listener in self._listeners:
            logger.debug('Configuring listener %s', listener)
            listener.configure(config)
        # save to file
        with open(path, 'wb') as configfile:
            logger.debug('Writing back configuration to %s', path)
            config.write(configfile)
        # as we store passwords in clear text, let's at least set correct file permissions
        logger.debug('Setting owner only file permissions to %s', path)
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)

    def run(self, polling_intervall=1):
        """
        Starts the polling loop that will run until receive common exit signals.

        :param polling_intervall: how often the connector polls data from haster in seconds (default: 1)
        :type  polling_intervall: int
        """
        try:
            for listener in self._listeners:
                logger.debug('Preparing listener %s', listener)
                listener.prepare()
            logger.info('Start listening for hamster activity...')
            while True:
                now = datetime.datetime.now().replace(microsecond=0)
                last = now - datetime.timedelta(seconds=polling_intervall)
                for fact in self.get_todays_facts():
                    if fact.start_time is not None and last <= fact.start_time < now:
                        logger.debug('Found a started task: %r', vars(fact))
                        for listener in self._listeners:
                            listener.on_fact_started(fact)
                    if fact.end_time is not None and last <= fact.end_time < now:
                        logger.debug('Found a stopped task: %r', vars(fact))
                        for listener in self._listeners:
                            listener.on_fact_stopped(fact)
                time.sleep(polling_intervall)
        except (KeyboardInterrupt, SystemExit):
            pass



from __future__ import absolute_import

import logging
import re

from hamster_bridge.listeners import HamsterListener


logger = logging.getLogger(__name__)


class RedmineHamsterListener(HamsterListener):
    """
    Redmine listener for hamster tasks,
    Tested with Redmine 2.5.1.stable

    Important: will only work with German or English installation!

    INFO: Unfortunately the Redmine API returns issue statuses in the currently set language.
          There is only the id and the name of the status.
          f.e. "New" has usually ID 1, but its name would be "Neu" in a German installation.
    """
    short_name = 'redmine'

    config_values = [
        ('server_url', lambda: raw_input('Root URL to the Redmine server [f.e. "http://redmine.example.org/"]\n')),
        ('api_key', lambda: raw_input('Your Redmine API access key.\n')),
        ('version', lambda: raw_input('The Redmine version number, e.g. 2.5.1\n')),
        ('auto_start', lambda: raw_input('Automatically start the issue when you start the task in hamster? [y/n]\n')),
        # FIXME still usable?
        ('verify_ssl', lambda: raw_input('Verify HTTPS/SSL connections? [y/n]\n')),
    ]

    # Redmine issue key is just a number
    issue_from_title = re.compile('([0-9]+)\ ')

    def __get_config(self, key):
        """
        Returns the config value with the given key.
        :param key: the key to get
        :type key: basestring
        :return: the config value
        :rtype: basestring
        """
        return self.config.get(self.short_name, key)

    def __init__(self):
        """
        Sets up the class be defining some internal variables.
        """
        # issue status dict for the default issue status
        self.__issue_status_default = None

        # issue status dict for the "in Work" status
        self.__issue_status_in_work = None

        # the redmine instance
        self.redmine = None

        # will store the activities
        self.__activities = {}

    def __get_issue_from_fact(self, fact):
        """
        Tries to find an issue matching the given fact.

        :param fact: the currently stopped fact
        :type fact: hamster.lib.stuff.Fact
        :returns: the issue or None if not found
        :rtype:
        """
        from redmine.exceptions import ResourceNotFoundError
        
        # iterate the possible issues, normally this should match exactly one...
        for possible_issue in self.issue_from_title.findall(fact.activity):
            try:
                return self.redmine.issue.get(possible_issue)
            except ResourceNotFoundError:
                return None

        return None

    def __filter_issue_statuses(self):
        """
        Filters the issue statuses for the relevant ones: the default and the status "In Work".
        """

        def find_default(element):
            """
            Filter function to find the default issue status.
            """
            return hasattr(element, 'is_default') and getattr(element, 'is_default', False)

        def find_in_work(element):
            """
            Filter function to find the in work status.
            """
            return element.name in [u'In Bearbeitung', u'In Work']

        # get the issue statuses
        issue_statuses = self.redmine.issue_status.all()

        # if none are found, raise
        if len(issue_statuses) == 0:
            logger.exception('Unable to fetch issue statuses! Not possible to proceed!')

        try:
            self.__issue_status_default = [item for item in issue_statuses if find_default(item)][0]
        except IndexError:
            logger.exception('Unable to find a single default issue status!')

        try:
            self.__issue_status_in_work = [item for item in issue_statuses if find_in_work(item)][0]
        except IndexError:
            logger.exception('Unable to find a single "In Work" issue status!')

    def __get_first_activity_id(self):
        """
        Returns the first activity that was retrieved.
        The first activity was marked with a True value.
        :return: the id of the first activity
        :rtype: int
        """
        return [key for key, value in self.__activities.items() if value[1]][0]

    def __get_activity_id(self, tags):
        """
        Returns an activity id if it can be resolved from the list of given tags.
        Otherwise the first found activity id is returned.
        :param tags: list of tags
        :type tags: list
        :return: the activity id
        :rtype: int
        """
        if len(tags) == 0:
            # this grabs the first activity from the dict
            return self.__get_first_activity_id()
        else:
            for activity_id, activity_value in self.__activities.viewitems():
                try:
                    next(tag for tag in tags if tag == activity_value[0])
                    return activity_id
                except StopIteration:
                    # if not found
                    continue

            # fallback if no tag matches
            return self.__get_first_activity_id()

    def prepare(self):
        """
        Prepares the listener by checking connectivity to configured Redmine instance.
        While doing so, grabs the issue statuses, too, used for on_fact_stopped.
        """
        from redmine import Redmine
        from redmine.exceptions import BaseRedmineError
        
        # setup the redmine instance
        self.redmine = Redmine(
            self.__get_config('server_url'),
            key=self.__get_config('api_key'),
            version=self.__get_config('version'),
            requests={
                'verify': True if self.__get_config('verify_ssl') == 'y' else False
            }
        )
        # fetch the possible activities for time entries
        time_entry_activities = self.redmine.enumeration.filter(resource='time_entry_activities')

        # only now the real http request is made, use this as connectivity check
        try:
            logger.info('### Available Redmine activities for using as tag value:')
            is_first = True
            for tea in time_entry_activities:
                self.__activities[tea.id] = (tea.name, is_first)
                is_first = False
                logger.info('### ' + tea.name)
        except (BaseRedmineError, IOError):
            logger.exception('Unable to communicate with redmine server. See error in the following output:')

        # fetch all available issue statuses and filter the default and in work ones as they are the only relevant statuses here
        self.__filter_issue_statuses()

    def on_fact_started(self, fact):
        """
        Called by HamsterBridge if a fact is started.
        Will try to start the appropriate Redmine issue if there is one.
        Uses the first found issue.

        :param fact: the currently stopped fact
        :type fact: hamster.lib.stuff.Fact
        """
        # if issue shall be auto started...
        if self.config.get(self.short_name, 'auto_start') == 'y':
            # fetch the issue from the hamster fact
            issue = self.__get_issue_from_fact(fact)

            # abort if no issue was found
            if not issue:
                logger.error('Unable to query issue for starting of hamster fact %s', fact.original_activity)
                return

            # if the issue is in the default state (aka the initial state), put it into work state
            if issue.status.id == self.__issue_status_default.id:
                logger.info('setting status to "In Work" for issue %d', issue.id)
                issue.status_id = self.__issue_status_in_work.id
                issue.save()

    def on_fact_stopped(self, fact):
        """
        Called by HamsterBridge if a fact is stopped.
        Will try to log the time to the appropriate Redmine issue if there is one.
        Uses the first found issue.

        :param fact: the currently stopped fact
        :type fact: hamster.lib.stuff.Fact
        """

        # fetch the issue from the hamster fact
        issue = self.__get_issue_from_fact(fact)

        # abort if no issue was found
        if not issue:
            logger.error('Unable to query issue for stopping of hamster fact %s', fact.original_activity)
            return

        # create the time entry
        self.redmine.time_entry.create(
            issue_id=issue.id,
            hours='%0.2f' % (fact.delta.total_seconds() / 3600.0),
            # grep the tags, convert to string (the values are dbus.String) and find an activity
            activity_id=self.__get_activity_id([str(tag) for tag in fact.tags]),
            comments=fact.description,
        )





class HamsterListener(object):

    short_name = None
    config_values = []

    def configure(self, config):
        self.config = config
        if self.short_name is not None and len(self.config_values) > 0:
            if not self.config.has_section(self.short_name):
                self.config.add_section(self.short_name)
            for config_key, config_value in self.config_values:
                if not self.config.has_option(self.short_name, config_key):
                    self.config.set(self.short_name, config_key, config_value())

    def prepare(self):
        pass

    def on_fact_started(self, fact):
        pass

    def on_fact_stopped(self, fact):
        pass


#!/usr/bin/env python
from setuptools import setup


setup(
    name='hamster-bridge',
    description='let your hamster log your work to your favorite bugtracker',
    version='0.5.1',
    author='Lars Kreisz',
    author_email='lars.kreisz@gmail.com',
    license='MIT',
    url='https://github.com/kraiz/hamster-bridge',
    extras_require={
        'redmine': ['python-redmine'],
    },
    packages=['hamster_bridge', 'hamster_bridge.listeners'],
    entry_points={'console_scripts': ['hamster-bridge = hamster_bridge:main']},
    long_description=open('README.rst').read() + '\n\n' + open('CHANGELOG.rst').read(),
    install_requires=['jira>=0.41']
)
